/// Represents a location in a file by line and column.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct Position
{
   /// The referenced line of the file.
   pub line: usize,

   /// The referenced column of the file.
   pub column: usize
}


/// Represents a range within a file.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct Span
{
   /// The start `Position` in the file.
   pub start: Position,

   /// The end `Position` in the file.
   pub end: Position
}



impl Position
{
   pub fn new(line: usize, column: usize) -> Self
   {
      Position { line, column }
   }
}

impl Default for Position
{
   fn default() -> Self
   {
      Position { line: 0, column: 0 }
   }
}

impl std::fmt::Display for Position
{
   fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result
   {
      write!(f, "Line {}, Column {}", self.line, self.column)
   }
}



impl Span
{
   pub fn new(start: Position, end: Position) -> Self
   {
      Span { start, end }
   }

   pub fn with_single(val: Position) -> Self
   {
      Span { start: val,
             end: val }
   }

   pub fn merge(a: Span, b: Span) -> Self
   {
      Span { start: a.start,
             end: b.end }
   }

   pub fn merge_with(&self, other: Span) -> Span {
        Span {
            start: self.start,
            end: other.end,
        }
    }
}

impl Default for Span
{
   fn default() -> Self
   {
      Span { start: Position::default(),
             end: Position::default() }
   }
}

impl std::fmt::Display for Span
{
   fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result
   {
      write!(f, "Span[{} -> {}]", self.start, self.end)
   }
}
